import sys
import os
from cffi import FFI

PATH = os.path.dirname(__file__)

with open(os.path.join(PATH, 'lib_build.h')) as hf:
    c_header = hf.read()
with open(os.path.join(PATH, 'lib_build.c')) as cf:
    c_source = cf.read()

ffi = FFI()
ffi.cdef(c_header)
ffi.set_source('_lib', c_source)

if __name__ == '__main__':
    ffi.compile()



"""
pyxattr and xattr have differing API, for example xattr assumes
that (like on OSX) attribute keys are valid UTF-8, while pyxattr
just passes through the raw bytestring.

This module provides compatibility for the pyxattr API.
"""

import sys

from .lib import (XATTR_NOFOLLOW, XATTR_CREATE, XATTR_REPLACE,
    XATTR_NOSECURITY, XATTR_MAXNAMELEN, XATTR_FINDERINFO_NAME,
    XATTR_RESOURCEFORK_NAME, _getxattr, _fgetxattr, _setxattr, _fsetxattr,
    _removexattr, _fremovexattr, _listxattr, _flistxattr)

__all__ = [
    "NS_SECURITY", "NS_USER", "NS_SYSTEM", "NS_TRUSTED",
    "getxattr", "get", "get_all", "setxattr", "set",
    "removexattr", "remove", "listxattr", "list"
]

NS_SECURITY = b"security"
NS_USER = b"user"
NS_SYSTEM = b"system"
NS_TRUSTED = b"trusted"

_NO_NS = object()

_fsencoding = sys.getfilesystemencoding()

def _call(item, name_func, fd_func, *args):
    if isinstance(item, int):
        return fd_func(item, *args)
    elif hasattr(item, 'fileno'):
        return fd_func(item.fileno(), *args)
    elif isinstance(item, bytes):
        return name_func(item, *args)
    elif isinstance(item, str):
        item = item.encode(_fsencoding)
        return name_func(item, *args)
    else:
        raise TypeError("argument must be string, int or file object")

def _add_ns(item, ns):
    if ns is None:
        raise TypeError("namespace must not be None")
    if ns == _NO_NS:
        return item
    return b'.'.join((ns, item))

def getxattr(item, attribute, nofollow=False):
    options = nofollow and XATTR_NOFOLLOW or 0
    return _call(item, _getxattr, _fgetxattr, attribute, 0, 0, options)

def get(item, name, nofollow=False, namespace=_NO_NS):
    name = _add_ns(name, namespace)
    return getxattr(item, name, nofollow=nofollow)

def get_all(item, nofollow=False, namespace=_NO_NS):
    if namespace is not None and namespace != _NO_NS:
        namespace += b'.'
    l = listxattr(item, nofollow=nofollow)
    result = []
    for name in l:
        try:
            if namespace is not None and namespace != _NO_NS:
                if not name.startswith(namespace):
                    continue
                result.append((name[len(namespace):],
                               getxattr(item, name, nofollow=nofollow)))
            else:
                result.append((name, getxattr(item, name, nofollow=nofollow)))
        except IOError:
            pass
    return result

def setxattr(item, name, value, flags=0, nofollow=False):
    options = nofollow and XATTR_NOFOLLOW or 0
    options |= flags
    return _call(item, _setxattr, _fsetxattr, name, value, 0, options)

def set(item, name, value, nofollow=False, flags=0, namespace=_NO_NS):
    name = _add_ns(name, namespace)
    return setxattr(item, name, value, flags=flags, nofollow=nofollow)

def removexattr(item, name, nofollow=False):
    options = nofollow and XATTR_NOFOLLOW or 0
    return _call(item, _removexattr, _fremovexattr, name, options)

def remove(item, name, nofollow=False, namespace=_NO_NS):
    name = _add_ns(name, namespace)
    return removexattr(item, name, nofollow=nofollow)

def listxattr(item, nofollow=False):
    options = nofollow and XATTR_NOFOLLOW or 0
    res = _call(item, _listxattr, _flistxattr, options).split(b'\x00')
    res.pop()
    return res

def list(item, nofollow=False, namespace=_NO_NS):
    if not namespace or namespace == _NO_NS:
        return listxattr(item, nofollow=nofollow)
    namespace += b'.'
    l = listxattr(item, nofollow=nofollow)
    result = []
    for name in l:
        if not name.startswith(namespace):
            continue
        result.append(name[len(namespace):])
    return result



##
# Copyright (c) 2007 Apple Inc.
#
# This is the MIT license. This software may also be distributed under the
# same terms as Python (the PSF license).
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
##

from __future__ import print_function

import sys
import os
import getopt
import zlib

import xattr


class NullsInString(Exception):
    """Nulls in string."""


def usage(e=None):
    if e:
        print(e)
        print("")

    name = os.path.basename(sys.argv[0])
    print("usage: %s [-slz] file [file ...]" % (name,))
    print("       %s -p [-slz] attr_name file [file ...]" % (name,))
    print("       %s -w [-sz] attr_name attr_value file [file ...]" % (name,))
    print("       %s -d [-s] attr_name file [file ...]" % (name,))
    print("       %s -c [-s] file [file ...]" % (name,))
    print("")
    print("The first form lists the names of all xattrs on the given file(s).")
    print("The second form (-p) prints the value of the xattr attr_name.")
    print("The third form (-w) sets the value of the xattr attr_name to attr_value.")
    print("The fourth form (-d) deletes the xattr attr_name.")
    print("The fifth form (-c) deletes (clears) all xattrs.")
    print("")
    print("options:")
    print("  -h: print this help")
    print("  -s: act on symbolic links themselves rather than their targets")
    print("  -l: print long format (attr_name: attr_value)")
    print("  -z: compress or decompress (if compressed) attribute value in zip format")

    if e:
        return 64
    else:
        return 0


_FILTER = u''.join([(len(ascii(chr(x))) == 3) and chr(x) or u'.' for x in range(256)])


def _dump(src, length=16):
    result = []
    for i in range(0, len(src), length):
        s = src[i:i+length]
        hexa = ' '.join(["%02X" % ord(x) for x in s])
        printable = s.translate(_FILTER)
        result.append("%04X   %-*s   %s\n" % (i, length*3, hexa, printable))
    return ''.join(result)


def main(argv=sys.argv):
    try:
        (optargs, args) = getopt.getopt(argv[1:], "hlpwdzsc", ["help"])
    except getopt.GetoptError as e:
        return usage(e)

    attr_name = None
    long_format = False
    read = False
    write = False
    delete = False
    clear = False
    nofollow = False
    compress = lambda x: x
    decompress = compress
    errors = []

    for opt, arg in optargs:
        if opt in ("-h", "--help"):
            return usage()
        elif opt == "-l":
            long_format = True
        elif opt == "-s":
            nofollow = True
        elif opt == "-p":
            read = True
            if write or delete or clear:
                return usage("-p not allowed with -w, -d or -c")
        elif opt == "-w":
            write = True
            if read or delete or clear:
                return usage("-w not allowed with -p, -d or -c")
        elif opt == "-d":
            delete = True
            if read or write or clear:
                return usage("-d not allowed with -p, -w or -c")
        elif opt == "-c":
            clear = True
            if read or write or delete:
                return usage("-c not allowed with -p, -w or -d")
        elif opt == "-z":
            compress = zlib.compress
            decompress = zlib.decompress

    if write or delete or clear:
        if long_format:
            return usage("-l not allowed with -w, -d or -c")

    if read or write or delete:
        if not args:
            return usage("No attr_name")
        attr_name = args.pop(0)

    if write:
        if not args:
            return usage("No attr_value")
        attr_value = args.pop(0).encode('utf-8')

    if len(args) == 0:
        return usage("No file")

    if len(args) > 1:
        multiple_files = True
    else:
        multiple_files = False

    options = 0
    if nofollow:
        options |= xattr.XATTR_NOFOLLOW

    for filename in args:
        def onError(e):
            errors.append(e)
            if not os.path.exists(filename):
                sys.stderr.write("No such file: %s\n" % (filename,))
            else:
                sys.stderr.write(str(e) + "\n")

        try:
            attrs = xattr.xattr(filename, options=options)
        except (IOError, OSError) as e:
            onError(e)
            continue

        if write:
            try:
                attrs[attr_name] = compress(attr_value)
            except (IOError, OSError) as e:
                onError(e)
                continue

        elif delete:
            try:
                del attrs[attr_name]
            except (IOError, OSError) as e:
                onError(e)
                continue
            except KeyError:
                onError("No such xattr: %s" % (attr_name,))
                continue

        elif clear:
            try:
                attrs.clear()
            except (IOError, OSError) as e:
                onError(e)
                continue

        else:
            try:
                if read:
                    attr_names = (attr_name,)
                else:
                    attr_names = attrs.keys()
            except (IOError, OSError) as e:
                onError(e)
                continue

            if multiple_files:
                file_prefix = "%s: " % (filename,)
            else:
                file_prefix = ""

            for attr_name in attr_names:
                should_dump = False
                try:
                    try:
                        attr_value = decompress(attrs[attr_name])
                    except zlib.error:
                        attr_value = attrs[attr_name]
                    try:
                        if b'\0' in attr_value:
                            # force dumping
                            raise NullsInString
                        attr_value = attr_value.decode('utf-8')
                    except (UnicodeDecodeError, NullsInString):
                        attr_value = attr_value.decode('latin-1')
                        should_dump = True
                except KeyError:
                    onError("%sNo such xattr: %s" % (file_prefix, attr_name))
                    continue

                if long_format:
                    if should_dump:
                        print("".join((file_prefix, "%s:" % (attr_name,))))
                        print(_dump(attr_value))
                    else:
                        print("".join((file_prefix, "%s: " % (attr_name,), attr_value)))
                else:
                    if read:
                        if should_dump:
                            if file_prefix:
                                print(file_prefix)
                            print(_dump(attr_value))
                        else:
                            print("".join((file_prefix, attr_value)))
                    else:
                        print("".join((file_prefix, attr_name)))

    return 1 if errors else 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))
